@@ -32,13 +32,19 @@
 #include "gsm-icon-names.h"
 #include "gsm-manager.h"
 #include "gsm-util.h"
+#include "gsm-shell-extensions.h"
 
 #define GSM_FAIL_WHALE_DIALOG_GET_PRIVATE(o)                                \
         (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSM_TYPE_FAIL_WHALE_DIALOG, GsmFailWhaleDialogPrivate))
 
 struct _GsmFailWhaleDialogPrivate
 {
+        gboolean debug_mode;
+        gboolean allow_logout;
         GdkRectangle geometry;
+        gboolean show_extensions;
+        GsmShellExtensions *extensions;
+        GtkWidget *extensions_box;
 };
 
 G_DEFINE_TYPE (GsmFailWhaleDialog, gsm_fail_whale_dialog, GTK_TYPE_WINDOW);
@@ -78,6 +84,9 @@ _window_move_resize_window (GsmFailWhaleDialog *window,
 
         g_assert (gtk_widget_get_realized (widget));
 
+        if (window->priv->debug_mode)
+                return;
+
         g_debug ("Move and/or resize window x=%d y=%d w=%d h=%d",
                  window->priv->geometry.x,
                  window->priv->geometry.y,
@@ -122,6 +131,20 @@ on_screen_size_changed (GdkScreen          *screen,
         gtk_widget_queue_resize (GTK_WIDGET (fail_dialog));
 }
 
+static void
+gsm_fail_whale_dialog_finalize (GObject *object)
+{
+        GsmFailWhaleDialog *fail_dialog = GSM_FAIL_WHALE_DIALOG (object);
+        GsmFailWhaleDialogPrivate *priv = fail_dialog->priv;
+
+        if (priv->extensions != NULL) {
+                g_object_unref (priv->extensions);
+                priv->extensions = NULL;
+        }
+
+        G_OBJECT_CLASS (gsm_fail_whale_dialog_parent_class)->finalize (object);
+}
+
 static void
 gsm_fail_whale_dialog_realize (GtkWidget *widget)
 {
@@ -221,8 +244,12 @@ gsm_fail_whale_dialog_get_preferred_height (GtkWidget *widget,
 static void
 gsm_fail_whale_dialog_class_init (GsmFailWhaleDialogClass *klass)
 {
+        GObjectClass   *object_class;
         GtkWidgetClass *widget_class;
 
+        object_class = G_OBJECT_CLASS (klass);
+        object_class->finalize = gsm_fail_whale_dialog_finalize;
+
         widget_class = GTK_WIDGET_CLASS (klass);
 
         widget_class->realize = gsm_fail_whale_dialog_realize;
@@ -237,16 +264,76 @@ static void
 on_logout_clicked (GtkWidget          *button,
                    GsmFailWhaleDialog *fail_dialog)
 {
-        gsm_manager_logout (gsm_manager_get (),
-                            GSM_MANAGER_LOGOUT_MODE_FORCE,
-                            NULL);
+        if (fail_dialog->priv->debug_mode) {
+                gtk_main_quit ();
+        } else {
+                gsm_manager_logout (gsm_manager_get (),
+                                    GSM_MANAGER_LOGOUT_MODE_FORCE,
+                                    NULL);
 
-        gtk_widget_destroy (GTK_WIDGET (fail_dialog));
+                gtk_widget_destroy (GTK_WIDGET (fail_dialog));
+        }
+}
+
+static void
+on_extension_switched (GObject            *object,
+                       GParamSpec         *spec,
+                       GsmFailWhaleDialog *fail_dialog)
+{
+        gchar *uuid;
+        gboolean enabled;
+
+        uuid = g_object_get_data (object, "uuid");
+        g_object_get (object, "active", &enabled, NULL);
+
+        gsm_shell_extensions_set_enabled (fail_dialog->priv->extensions,
+                                          uuid,
+                                          enabled);
+}
+
+static void
+build_extension_toggle (GsmShellExtensions *extensions,
+                        GsmShellExtension  *extension,
+                        gpointer            data)
+{
+        GsmFailWhaleDialog *fail_dialog = GSM_FAIL_WHALE_DIALOG (data);
+        GtkWidget *hbox;
+        GtkWidget *label;
+        GtkWidget *sw;
+
+        hbox = gtk_hbox_new (FALSE, 0);
+        gtk_widget_set_size_request (hbox, 600, -1);
+        gtk_widget_set_halign (hbox, GTK_ALIGN_CENTER);
+        gtk_widget_show (hbox);
+        gtk_box_pack_start (GTK_BOX (fail_dialog->priv->extensions_box),
+                            hbox, FALSE, FALSE, 0);
+
+        label = gtk_label_new (gsm_shell_extension_get_name (extension));
+
+        gtk_widget_show (label);
+        gtk_widget_set_halign (label, GTK_ALIGN_START);
+        gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
+
+        sw = gtk_switch_new ();
+        gtk_switch_set_active (GTK_SWITCH (sw),
+                               gsm_shell_extension_get_is_enabled (extension));
+
+        g_object_set_data (G_OBJECT (sw),
+                           "uuid",
+                           gsm_shell_extension_get_uuid (extension));
+        g_signal_connect (G_OBJECT (sw),
+                          "notify::active",
+                          G_CALLBACK (on_extension_switched),
+                          fail_dialog);
+        gtk_widget_set_halign (label, GTK_ALIGN_END);
+        gtk_widget_show (sw);
+        gtk_box_pack_end (GTK_BOX (hbox), sw, FALSE, FALSE, 0);
 }
 
 static void
 setup_window (GsmFailWhaleDialog *fail_dialog)
 {
+        GsmFailWhaleDialogPrivate *priv;
         GtkWidget *alignment;
         GtkWidget *box;
         GtkWidget *image;
@@ -254,8 +341,24 @@ setup_window (GsmFailWhaleDialog *fail_dialog)
         GtkWidget *message_label;
         GtkWidget *button_box;
         GtkWidget *button;
+        GtkWidget *extensions_scroll;
+        GtkWidget *extensions_vbox;
         char *markup;
 
+        priv = fail_dialog->priv;
+
+        gtk_window_set_title (GTK_WINDOW (fail_dialog), "");
+        gtk_window_set_icon_name (GTK_WINDOW (fail_dialog), GSM_ICON_COMPUTER_FAIL);
+
+        if (!fail_dialog->priv->debug_mode) {
+                gtk_window_set_skip_taskbar_hint (GTK_WINDOW (fail_dialog), TRUE);
+                gtk_window_set_keep_above (GTK_WINDOW (fail_dialog), TRUE);
+                gtk_window_stick (GTK_WINDOW (fail_dialog));
+                gtk_window_set_position (GTK_WINDOW (fail_dialog), GTK_WIN_POS_CENTER_ALWAYS);
+                /* only works if there is a window manager which is unlikely */
+                gtk_window_fullscreen (GTK_WINDOW (fail_dialog));
+        }
+
         alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
         gtk_widget_show (alignment);
         gtk_container_add (GTK_CONTAINER (fail_dialog), alignment);
@@ -277,47 +380,66 @@ setup_window (GsmFailWhaleDialog *fail_dialog)
         gtk_widget_show (label);
         gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
 
-        /* make this changable at some point */
-        message_label = gtk_label_new (_("A problem has occurred and the system can't recover.\nPlease log out and try again."));
+        if (!priv->allow_logout)
+                message_label = gtk_label_new (_("A problem has occurred and the system can't recover. Please contact a system administrator"));
+        else if (priv->show_extensions && gsm_shell_extensions_n_extensions (priv->extensions) > 0)
+                message_label = gtk_label_new (_("A problem has occurred and the system can't recover. Some of the extensions below may have caused this.\nPlease try disabling some of these, and then log out and try again."));
+        else
+                message_label = gtk_label_new (_("A problem has occurred and the system can't recover.\nPlease log out and try again."));
+
         gtk_label_set_justify (GTK_LABEL (message_label), GTK_JUSTIFY_CENTER);
         gtk_label_set_line_wrap (GTK_LABEL (message_label), TRUE);
         gtk_widget_show (message_label);
         gtk_box_pack_start (GTK_BOX (box),
                             message_label, FALSE, FALSE, 0);
 
+        if (priv->allow_logout && priv->show_extensions) {
+                extensions_vbox = gtk_vbox_new (TRUE, 6);
+                gtk_widget_show (extensions_vbox);
+                fail_dialog->priv->extensions_box = extensions_vbox;
+
+                extensions_scroll = gtk_scrolled_window_new (NULL, NULL);
+                gtk_widget_show (extensions_scroll);
+
+                gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (extensions_scroll),
+                                                       extensions_vbox);
+
+                gtk_box_pack_start (GTK_BOX (box),
+                                    extensions_scroll, FALSE, FALSE, 0);
+
+                gsm_shell_extensions_foreach (priv->extensions,
+                                              build_extension_toggle,
+                                              (gpointer) fail_dialog);
+        }
+
         button_box = gtk_hbutton_box_new ();
         gtk_container_set_border_width (GTK_CONTAINER (button_box), 20);
         gtk_widget_show (button_box);
         gtk_box_pack_end (GTK_BOX (box),
                           button_box, FALSE, FALSE, 0);
 
-        button = gtk_button_new_with_mnemonic (_("_Log Out"));
-        gtk_widget_show (button);
-        gtk_box_pack_end (GTK_BOX (button_box),
-                          button, FALSE, FALSE, 0);
-        g_signal_connect (button, "clicked",
-                          G_CALLBACK (on_logout_clicked), fail_dialog);
+        if (priv->allow_logout) {
+                button = gtk_button_new_with_mnemonic (_("_Log Out"));
+                gtk_widget_show (button);
+                gtk_box_pack_end (GTK_BOX (button_box),
+                                  button, FALSE, FALSE, 0);
+                g_signal_connect (button, "clicked",
+                                  G_CALLBACK (on_logout_clicked), fail_dialog);
+        }
 }
 
 static void
 gsm_fail_whale_dialog_init (GsmFailWhaleDialog *fail_dialog)
 {
         fail_dialog->priv = GSM_FAIL_WHALE_DIALOG_GET_PRIVATE (fail_dialog);
-
-        gtk_window_set_skip_taskbar_hint (GTK_WINDOW (fail_dialog), TRUE);
-        gtk_window_set_keep_above (GTK_WINDOW (fail_dialog), TRUE);
-        gtk_window_stick (GTK_WINDOW (fail_dialog));
-        gtk_window_set_title (GTK_WINDOW (fail_dialog), "");
-        gtk_window_set_icon_name (GTK_WINDOW (fail_dialog), GSM_ICON_COMPUTER_FAIL);
-        gtk_window_set_position (GTK_WINDOW (fail_dialog), GTK_WIN_POS_CENTER_ALWAYS);
-        /* only works if there is a window manager which is unlikely */
-        gtk_window_fullscreen (GTK_WINDOW (fail_dialog));
-
-        setup_window (fail_dialog);
+        fail_dialog->priv->extensions = g_object_new (GSM_TYPE_SHELL_EXTENSIONS, NULL);
 }
 
 void
-gsm_fail_whale_dialog_we_failed (void)
+gsm_fail_whale_dialog_we_failed (gboolean debug_mode,
+                                 gboolean allow_logout,
+                                 gboolean show_extensions)
+
 {
         static GsmFailWhaleDialog *current_dialog = NULL;
         GsmFailWhaleDialog        *fail_dialog;
@@ -327,6 +449,10 @@ gsm_fail_whale_dialog_we_failed (void)
         }
 
         fail_dialog = g_object_new (GSM_TYPE_FAIL_WHALE_DIALOG, NULL);
+        fail_dialog->priv->debug_mode = debug_mode;
+        fail_dialog->priv->allow_logout = allow_logout;
+        fail_dialog->priv->show_extensions = show_extensions;
+        setup_window (fail_dialog);
 
         current_dialog = fail_dialog;
         g_signal_connect (current_dialog,
